%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Bacteria biofilm simulation
% Published in BMC Systems Biology
% Written by Poonam Phalak, Jin Chen and Michael A. Henson
% December, 2015

function [lb,ub] = RHS( t,y,INFO )

% This subroutine updates the upper and lower bounds for the fluxes in the
% exID arrays in main. The output should be two matrices, lb and ub. The lb matrix
% contains the lower bounds for exID{i} in the ith row in the same order as
% exID. The same is true for the upper bounds in the ub matrix.
% Infinity can be used for unconstrained variables, however, it should be 
% fixed for all time. 

nmodel = INFO.nmodel;
N = INFO.N;
ns = INFO.ns;
param = INFO.param;

% Set substrate uptake parameters

xv_glcm = param(1);
xKglc   = param(2);
xv_o2m = param(3);
xKo2   = param(4);
xv_acem = param(5);
xKace   = param(6);
xv_succm = param(7);
xKsucc   = param(8);

for i = 1:N+1
    
    % Set state variables
   
    xbm(i) = y(1+(i-1)*ns);     % PA biomass concentration, g/L
    glc(i) = y(2+(i-1)*ns);     % Glucose concentration, mmol/L
    o2(i) = y(3+(i-1)*ns);      % Oxygen concentration, mmol/L
    ace(i) = y(4+(i-1)*ns);     % Acetate concentration, mol/L
    succ(i) = y(5+(i-1)*ns);    % Succinate concentration, mol/L
    
    % Compute substrate uptake bounds
    
    xglcf(i) = xv_glcm*max([glc(i) 0])/(xKglc + glc(i));       % PA glucose uptake, mmol/g/h
    xo2f(i) = xv_o2m*max([o2(i) 0])/(xKo2 + o2(i));            % PA oxygen uptake, mmol/g/h   
    xacef(i) = xv_acem*max([ace(i) 0])/(xKace + ace(i));       % PA acetate uptake, mmol/g/h
    xsuccf(i) = xv_succm*max([succ(i) 0])/(xKsucc + succ(i));  % PA succinate uptake, mmol/g/h

    % Set biomass and substrate uptake bounds
    
    lb(i,1) = 0;                % PA biomass, 1/h
    ub(i,1) = Inf;
    
    lb(i,2) = -xglcf(i);        % PA glucose uptake bound, mmol/g/h
    ub(i,2) = Inf;
    
    lb(i,3) = -xo2f(i);         % PA oxygen uptake bound, mmol/g/h
    ub(i,3) = Inf;
    
    lb(i,4) = -xacef(i);        % PA acetate uptake bound, mmol/g/h
    ub(i,4) = Inf;
    
    lb(i,5) = -xsuccf(i);       % PA succinate uptake bound, mmol/g/h
    ub(i,5) = Inf;
    
 end
end

