%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Bacteria biofilm simulation - main program
% Published in BMC Systems Biology
% Written by Poonam Phalak, Jin Chen and Michael A. Henson
% December, 2015

% Clear workspace

clear all

% Specify number of discretization points

nspecies = 1;
npoints = 50;
nmodel = nspecies*npoints;
INFO.nmodel = nmodel;

% Load models. These should be .mat files generated by the COBRA toolbox. 
% When generating these files using the COBRA toolbox, a big number is used
% as infinity. This number should be fed to the DB vector (Default bound).
    
load model.mat
for i = 1:npoints
    model{i} = pa;
    DB(i) = 1000;
    end
INFO.DB = DB;

%% exID array
% You can either search the reaction names by name or provide them directly
% in the exID array.
% RxnNames = {'EX_glc(e)', 'EX_ac(e)', 'biomass'};
% for i = 1:length(RxnNames)
%    [a,exID(i)] = ismember(RxnNames(i),model.rxns);
% end

for i = 1:npoints
    exID{i}=[88 135 177 95 197];
    end
INFO.exID = exID;

% Lower bounds and upper bounds for these reactions should be provided in
% the RHS code. 

% This codes solves the LPs in standard form. Bounds on exchange fluxes in 
% the exID array can be modified directly on the first 2*n rows where n is 
% the number of exchange fluxes. Order will be lower bound, upper bound, 
% lower bound, upper bound in the same order as exID. 
%
% NOTE: All bounds on fluxes in the exID arrays are relaxed to -Inf and 
% + Inf. These bounds need to be updated if needed in the RHS file.

%% Cost vectors
% Usually the first cost vector will be biomass maximization, but it can
% be any other objective. The CPLEX objects will minimize by default. 
% Report only nonzero elements. 
% The structure should be:
% C{model} = struct
% Element C{k}(i) of C, is the cost structure i for model k. 
% C{k}(i).sense = +1 for minimize, or -1 for maximize.
% C{k}(i).rxns = array containing the reactions in this objective. 
% C{k}(i).wts = array containing coefficients for reactions reported in 
% rxns. Both arrays should have the same length. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% Cost{k}(i) = 3*v_144 + v_832 - v_931 (fluxes for model k). 
% This cost vector will be either maximized or minimized depending on the
% value of C{k}(i).sense.

% In SBML files, usually production fluxes are positive and uptake fluxes
% are negative. Keep in mind that maximizing a negative flux implies 
% minimizing its absolute value.
% Different models can have different number of objectives. 

minim = 1;
maxim = -1;

for i=1:npoints
    
    % Maximize PA growth
    C{i}(1).sense = maxim;
    C{i}(1).rxns = [88];
    C{i}(1).wts = [1];
    
    % Minimize PA acetate synthesis 
    C{i}(2).sense = minim;
    C{i}(2).rxns = [95];
    C{i}(2).wts = [1];
    
    % Minimize PA succinate synthesis
    C{i}(3).sense = minim;
    C{i}(3).rxns = [197];
    C{i}(3).wts = [1];
    
    % Maximize PA glucose uptake
    C{i}(4).sense = minim;
    C{i}(4).rxns = [135];
    C{i}(4).wts = [1];
    
    % Maximize PA oxygen uptake
    C{i}(5).sense = minim;
    C{i}(5).rxns = [177];
    C{i}(5).wts = [1];
    
    
end
INFO.C = C;    

% Set number of discretization points and state variables

N = npoints-1;
ns = 5;
INFO.N = N;
INFO.ns = ns;

% Time of simulation

tspan = [0,200];

% Biofilm conditions and parameters

L = 50;                   % Biofilm thickness, micron
x0 = 1;                   % PA initial biomass concentration, g/L
glcb = 7.5;               % Exudate glucose concentration, mmol/L
o2L = 0.21;               % Air oxygen concentration, mmol/L
aceb = 0;                 % Exudate acetate concentration, mmol/L
succb = 0;                % Exudate succinate concentration, mmol/L
glcdiff = 9.4e-6;         % Aqueous glucose diffusion coefficient, cm^2/s
o2diff =  26.8e-6;        % Aqueous oxygen diffusion coefficient, cm^2/s
acediff = 16.2e-6;        % Aqueous acetate diffusion coefficient, cm^2/s
succdiff = 12.6e-6;       % Aqueous succinate diffusion coefficient, cm^2/s
glcmt = 2e-4;             % Glucose mass transfer coefficient, cm/s
o2mt = 2e-2;              % Oxygen mass transfer coefficient, cm/s
acemt = 2e-4;             % Acetate mass transfer coefficient, cm/s
succmt = 2e-4;            % Succinate mass transfer coefficient, cm/s
kpad = 0.01;              % PA death rate constant, 1/h
xmax = 200;               % Combined maximum biomass concentration, g/L
zs = L/N;                 % Spatial step size, micron

% Set parameter vector

condit = [L,glcdiff,o2diff,acediff,succdiff,glcmt,o2mt,acemt,...
    succmt,xmax,N,ns,zs,glcb,o2L,aceb,succb,kpad];
INFO.condit=condit;

% Uptake scenario and parameters

uptake = [1       % PA glucose uptake
          1       % PA oxygen uptake
          0       % PA acetate uptake
          0       % PA succinate uptake
         ];

param = [
     10              % PA maximum glucose uptake rate, mmol/g/h 
     0.5             % PA glucose uptake saturation constant, mmol/L 
     20              % PA maximum oxygen uptake rate, mmol/g/h  
     0.003           % PA oxygen uptake saturation constant, mmol/L 
     10              % PA maximum acetate uptake rate, mmol/g/h 
     0.5             % PA acetate uptake saturation constant, mmol/L
     10              % PA maximum succinate uptake rate, mmol/g/h
     0.5             % PA succinate uptake saturation constant, mmol/L
    ];

param(1:2:8) = uptake(1:4).*param(1:2:8);
INFO.param=param;

% Specify initial conditions for each discretization point
% Initial concentrations of PA biomass (g/L), SA biomass (g/L), glucose (mmol/L),
% oxygen (mmol/L), acetate (mmol/L), succinate (mmol/L), lactate (mmol/L), inhibitor (mmol/L) 

yo = [];
for i = 1:N+1
    yz = [x0,glcb,o2L,aceb,succb];
    yo = [yo yz];
end

%yo = pa_sa_ic_N25_L300;

% Add initial condition for penalty variable

yo = [yo 0];
Y0 = yo;

% CPLEX Objects construction parameters

INFO.LPsolver = 1; % CPLEX = 0, Gurobi = 1.
                   % CPLEX works equally fine with both methods.
                   % Gurobi seems to work better with Method = 1, and 
                   % Mosek with Method = 0.
INFO.tol = 1E-9; % Feasibility, optimality and convergence tolerance for Cplex (tol>=1E-9). 
                 % It is recommended it is at least 2 orders of magnitude
                 % tighter than the integrator tolerance. 
                 % If problems with infeasibility messages, tighten this
                 % tolerance.
INFO.tolPh1 = INFO.tol; % Tolerance to determine if a solution to phaseI equals zero.
                        % It is recommended to be the same as INFO.tol. 
INFO.tolevt = 5*INFO.tol; % Tolerance for event detection. Has to be greater
                          % than INFO.tol.

% You can modify the integration tolerances here.
% If some of the flows become negative after running the simulation once
% you can add the 'Nonnegative' option.
% Include function to compute non-zero Jacobian entries.

NN = 1:length(Y0);
options = odeset('AbsTol',1E-6,'RelTol',1E-6,'Nonnegative',NN,...
    'Events',@evts,'JPattern',jac_pa_biofilm(N,ns));

% INFO: You can use the INFO struct to pass parameters. Don't use any of 
% the names already declared or: INFO.t (carries time information), 
% INFO.ncost, INFO.lexID, INFO.LlexID, INFO.lbct, INFO.ubct, INFO.sense,
% INFO.b, INFO.pair. 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[model,INFO] = ModelSetupM(model,Y0,INFO);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if INFO.LPsolver == 0
    [INFO] = LexicographicOpt(model,INFO);
elseif INFO.LPsolver == 1
    [INFO] = LexicographicOptG(model,INFO);
else
    display('Solver not currently supported.');
end

tint = 0;
TF = [];
YF = [];
while tint<tspan(2)
    
% Look at MATLAB documentation if you want to change solver.
% ode15s is more or less accurate for stiff problems.

    [T,Y] = ode15s(@odes_pa_sa_biofilm,tspan,Y0,options,INFO);
    TF = [TF;T];
    YF = [YF;Y];
    tint = T(end);
    tspan = [tint,tspan(2)];
    Y0 = Y(end,:);
    if tint == tspan(2)
        break;
    end
    
% Determine model with basis change

    value = evts(tint,Y0,INFO);
    [jjj,j] = min(value);
    ct = 0;
    k = 0;
    while j>ct
        k = k + 1;
        ct = ct + size(model{k}.A,1);
    end
    INFO.flagbasis = k;
    fprintf('Basis change at time %d. \n',tint);
    
% Update b vector 

[INFO] = bupdate(tint,Y0,INFO);

% Perform lexicographic optimization

if INFO.LPsolver == 0
    [INFO] = LexicographicOpt(model,INFO);
elseif INFO.LPsolver == 1
    [INFO] = LexicographicOptG(model,INFO);
else
    display('Solver not currently supported.');
end
end

T = TF;
Y = YF;

% Print variables at the tissue-biofilm and biofilm-air interfaces at the
% final time.

Y(end,1:ns+1)'
Y(end,ns*N+1:ns*N+ns+1)'
       
%% Plotting

% Plot time profiles at tissue-biofilm interface

figure(1)
subplot(3,2,1)
plot(T,Y(:,1))
xlabel('Time [h]')
ylabel('PA Biomass [g/L]')

subplot(3,2,2)
plot(T,Y(:,2))
xlabel('Time [h]')
ylabel('Glucose [mmol/L]')

subplot(3,2,3)
plot(T,Y(:,3))
xlabel('Time [h]')
ylabel('Oxygen [mmol/L]')

subplot(3,2,4)
plot(T,Y(:,4))
xlabel('Time [h]')
ylabel('Acetate [mmol/L]')

subplot(3,2,5)
plot(T,Y(:,5))
xlabel('Time [h]')
ylabel('Succinate [mmol/L]')


% Plot time profiles in middle of biofilm

mz = round(N/2);

figure(2)
subplot(3,2,1)
plot(T,Y(:,ns*mz+1))
xlabel('Time [h]')
ylabel('PA Biomass [g/L]')

subplot(3,2,2)
plot(T,Y(:,ns*mz+2))
xlabel('Time [h]')
ylabel('Glucose [mmol/L]')

subplot(3,2,3)
plot(T,Y(:,ns*mz+3))
xlabel('Time [h]')
ylabel('Oxygen [mmol/L]')

subplot(3,2,4)
plot(T,Y(:,ns*mz+4))
xlabel('Time [h]')
ylabel('Acetate [mmol/L]')

subplot(3,2,5)
plot(T,Y(:,ns*mz+5))
xlabel('Time [h]')
ylabel('Succinate [mmol/L]')

% Plot time profiles at biofilm-air interface

figure(3)
subplot(3,2,1)
plot(T,Y(:,ns*N+1))
xlabel('Time [h]')
ylabel('PA Biomass [g/L]')

subplot(3,2,2)
plot(T,Y(:,ns*N+2))
xlabel('Time [h]')
ylabel('Glucose [mmol/L]')

subplot(3,2,3)
plot(T,Y(:,ns*N+3))
xlabel('Time [h]')
ylabel('Oxygen [mmol/L]')

subplot(3,2,4)
plot(T,Y(:,ns*N+4))
xlabel('Time [h]')
ylabel('Acetate [mmol/L]')

subplot(3,2,5)
plot(T,Y(:,ns*N+5))
xlabel('Time [h]')
ylabel('Succinate [mmol/L]')


% Plot spatial profiles at final time

for i=1:N+1
    zloc(i) = (i-1)*zs;
    xbm(i) = Y(end,ns*(i-1)+1);
    glc(i) = Y(end,ns*(i-1)+2);
    o2(i) = Y(end,ns*(i-1)+3);
    ace(i) = Y(end,ns*(i-1)+4);
    succ(i) = Y(end,ns*(i-1)+5);
end

figure(4)
subplot(3,2,1)
plot(zloc,xbm)
xlabel('Location (micron)')
ylabel('PA biomass [g/L]')

subplot(3,2,2)
plot(zloc,glc)
xlabel('Location (micron)')
ylabel('Glucose [mmol/L]')

subplot(3,2,3)
plot(zloc,o2)
xlabel('Location (micron)')
ylabel('Oxygen [mmol/L]')

subplot(3,2,4)
plot(zloc,ace)
xlabel('Location (micron)')
ylabel('Acetate [mmol/L]')

subplot(3,2,5)
plot(zloc,succ)
xlabel('Location (micron)')
ylabel('Succinate [mmol/L]')
