%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Bacteria biofilm simulation
% Published in BMC Systems Biology
% Written by Poonam Phalak, Jin Chen and Michael A. Henson
% December, 2015

function dy = odes_pa_sa_biofilm(t, y, INFO)

t

nmodel = INFO.nmodel;
N = INFO.N;
ns = INFO.ns;
param = INFO.param;
condit = INFO.condit;

% Set conditions

L = condit(1);
glcdiffa = 3600*(1e6)^2/(100)^2*condit(2);    % Aqueous glucose diffusion coefficient, micron^2/h
o2diffa = 3600*(1e6)^2/(100)^2*condit(3);     % Aqueous oxygen diffusion coefficient, micron^2/h
acediffa = 3600*(1e6)^2/(100)^2*condit(4);    % Aqueous acetate diffusion coefficient, micron^2/h
succdiffa = 3600*(1e6)^2/(100)^2*condit(5);   % Aqueous succinate diffusion coefficient, micron^2/h
glcmt = 3600*1e6/100*condit(6);               % Glucose mass transfer coefficient, micron/h
o2mt = 3600*1e6/100*condit(7);               % Oxygen mass transfer coefficient, micron/h
acemt = 3600*1e6/100*condit(8);              % Acetate mass transfer coefficient, micron/h
succmt = 3600*1e6/100*condit(9);             % Succinate mass transfer coefficient, micron/h
xmax = condit(10);
N = condit(11);
ns = condit(12);
zs = condit(13);
glcb = condit(14);
o2L = condit(15);
aceb = condit(16);
succb = condit(17);
kpad = condit(18);

% Set extracellular state variables

for i=1:N+1
    xbm(i) = y(1+(i-1)*ns);     % PA biomass concentration, g/L
    glc(i) = y(2+(i-1)*ns);     % Glucose concentration, mmol/L
    o2(i) = y(3+(i-1)*ns);      % Oxygen concentration, mmol/L
    ace(i) = y(4+(i-1)*ns);     % Acetate concentration, mol/L
    succ(i) = y(5+(i-1)*ns);    % Succinate concentration, mol/L
  
end

% The elements of the flux matrix have the sign given to them by the
% coefficients in the Cost vector in main. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% flux(k,i) = 3*v_144 + v_832 - v_931 
% The penalty is an array containing the feasibility LP problem optimal
% objective function value for each model. 

% Update bounds and solve for fluxes

[flux,penalty] = solveModel(t,y,INFO);

for i=1:N+1
    
% Set exchange fluxes
    
    xmu(i) = flux(i,1);           % PA growth rate, h-1
    xvace(i) = flux(i,2);         % PA acetate secretion rate, mmol/g/h
    xvsucc(i) = flux(i,3);        % PA succinate secretion rate, mmol/g/h
    xvglc(i) = flux(i,4);         % PA glucose uptake rate, mmol/g/h
    xvo2(i) = flux(i,5);          % PA oxygen uptake rate, mmol/g/h
    
    if penalty(i) > 1e-6
        xmu(i) = 0;
        xvace(i) = 0;
        xvsucc(i) = 0;
    end
    
% Calculate biofilm diffusion coefficients

    xtot(i) = xbm(i);
    glcdiff(i) = glcdiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    o2diff(i) = o2diffa*(1-0.19*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    acediff(i) = acediffa*(1-0.36*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    succdiff(i) = succdiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
  
end

% Set lower boundary conditions
        
xbm0 = xbm(2);
glc0 = glc(2)+2*glcmt*zs*(glcb-glc(1))/glcdiff(1);
o20 = o2(2)+2*o2mt*zs*(0.05-o2(1))/o2diff(1);
ace0 = ace(2)+2*acemt*zs*(aceb-ace(1))/acediff(1);
succ0 = succ(2)+2*succmt*zs*(succb-succ(1))/succdiff(1);
       
% Set upper boundary conditions
        
xbmf = xbm(N); 
glcf = glc(N);
o2f = o2(N)+2*o2mt*zs*(o2L-o2(N+1))/o2diff(N+1);
acef = ace(N);
succf = succ(N);

% Evaluate equations

dy = [];

for i=1:N+1
    
% Compute second order derivatives

    if i==1
        xbmd2(i) = (xbm(i+1)-2*xbm(i)+xbm0)/zs^2;
        glcd2(i) = (glc(i+1)-2*glc(i)+glc0)/zs^2;
        o2d2(i) = (o2(i+1)-2*o2(i)+o20)/zs^2;
        aced2(i) = (ace(i+1)-2*ace(i)+ace0)/zs^2;
        succd2(i) = (succ(i+1)-2*succ(i)+succ0)/zs^2;

                 
    elseif i==N+1
        xbmd2(i) = (xbmf-2*xbm(i)+xbm(i-1))/zs^2;
        glcd2(i) = (glcf-2*glc(i)+glc(i-1))/zs^2;
        o2d2(i) = (o2f-2*o2(i)+o2(i-1))/zs^2;
        aced2(i) = (acef-2*ace(i)+ace(i-1))/zs^2;
        succd2(i) = (succf-2*succ(i)+succ(i-1))/zs^2;

    else
        xbmd2(i) = (xbm(i+1)-2*xbm(i)+xbm(i-1))/zs^2;
        glcd2(i) = (glc(i+1)-2*glc(i)+glc(i-1))/zs^2;
        o2d2(i) = (o2(i+1)-2*o2(i)+o2(i-1))/zs^2;
        aced2(i) = (ace(i+1)-2*ace(i)+ace(i-1))/zs^2;
        succd2(i) = (succ(i+1)-2*succ(i)+succ(i-1))/zs^2;
       
    end
    
    % Evaluate RHS of ODEs
    
    dxbm(i) = xmu(i)*xbm(i)*(1-xtot(i)/xmax)-kpad*xbm(i);
    dglc(i) = xvglc(i)*xbm(i)+glcdiff(i)*glcd2(i);
    do2(i) = xvo2(i)*xbm(i)+o2diff(i)*o2d2(i);
    dace(i) = xvace(i)*xbm(i)+acediff(i)*aced2(i);
    dsucc(i) = xvsucc(i)*xbm(i)+succdiff(i)*succd2(i);
    
    dy = [dy dxbm(i) dglc(i) do2(i) dace(i) dsucc(i)];

end

dylast=0;
for i=1:nmodel
    dylast = dylast + penalty(i);
end
dy = [dy dylast];
dy = dy';
end
